import os

import torch
from torch_geometric.data import InMemoryDataset, Data


class GNNDataset(InMemoryDataset):
    """Create graph dataset from a numpy array"""

    def __init__(
        self,
        root,
        train_data,
        truth,
        processed_filename,
        n_events=1000,
        transform=None,
        pre_transform=None,
    ):
        """
        Args:
            root (str): Root directory where the dataset should be saved.
            raw_filename (str): Name of the data file.
            processed_filename (str): Name of the processed file.
            n_events (int): Number of events to be processed.
            features (list): List of feature columns to be used in the graph.
            truth (list): List of truth columns.
            transform (callable, optional): A function/transform that takes in an
                :obj:`torch_geometric.data.Data` object and returns a transformed
                version. The data object will be transformed before every access.
            pre_transform (callable, optional): A function/transform that takes in an
                :obj:`torch_geometric.data.Data` object and returns a transformed
                version. The data object will be transformed before being saved to
                disk.
        """
        self.train_data = train_data
        self.truth = truth

        self._processed_file_names = processed_filename
        self._root = root
        super().__init__(root, transform, pre_transform)
        self.data, self.slices = torch.load(self.processed_paths[0])

    @property
    def raw_dir(self):
        pass

    @property
    def raw_paths(self):
        pass

    @property
    def raw_file_names(self):
        pass

    @property
    def processed_dir(self):
        return os.path.join(self._root, "processed")

    @property
    def processed_file_names(self):
        return [self._processed_file_names]

    @property
    def processed_paths(self):
        return [
            os.path.join(self.processed_dir, processed_filename)
            for processed_filename in self.processed_file_names
        ]

    def process(self):
        total_events = 0
        data_list = []

        for i in range(len(self.train_data)):
            node_feats = self.train_data[i]
            ground_truth = self.truth[i]

            graph = Data(x=node_feats, y=ground_truth)
            data_list.append(graph)

        data, slices = self.collate(data_list)
        torch.save((data, slices), self.processed_paths[0])
